# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A command launcher launches a list of commands on a cluster; implement your own
launcher to add support for your cluster. We've provided an example launcher
which runs all commands serially on the local machine.
"""

import subprocess
import time
import torch
import os

def local_launcher(commands):
    """Launch commands serially on the local machine."""
    for cmd in commands:
        print('[starting]: ', cmd)
        subprocess.call(cmd, shell=True)
        print('[done]: ', cmd)

def dummy_launcher(commands):
    """
    Doesn't run anything; instead, prints each command.
    Useful for testing.
    """
    with open(f"dummy_sweep_{time.time()}.sh", "w") as f:
        f.write("#!/bin/bash\ngpu=\n\n")
        for cmd in commands:
            f.write(f"CUDA_VISIBLE_DEVICES=${{gpu}} {cmd}\n")
            print(f'Dummy launcher: {cmd}')

def dummy_multi_files_launcher(commands):
    """
    Doesn't run anything; instead, prints each command.
    Useful for testing.
    """
    with open("dummy_sweep_0.sh", "w") as f:
        f.write("#!/bin/bash\ngpu=0\n\n")
    with open("dummy_sweep_1.sh", "w") as f:
        f.write("#!/bin/bash\ngpu=1\n\n")
    with open("dummy_sweep_2.sh", "w") as f:
        f.write("#!/bin/bash\ngpu=2\n\n")

    for i, cmd in enumerate(commands):
        if i < len(commands)//3:
            file_id = 0
        elif i < (len(commands)//3) * 2:
            file_id = 1
        else:
            file_id = 2
        with open(f"dummy_sweep_{file_id}.sh", "a") as f:
            f.write(f"CUDA_VISIBLE_DEVICES=${{gpu}} {cmd}\n")
            print(f'Dummy launcher: {cmd}')

def multi_gpu_launcher(commands):
    """
    Launch commands on the local machine, using all GPUs in parallel.
    """
    print('WARNING: using experimental multi_gpu_launcher.')
    try:
        # Get list of GPUs from env, split by ',' and remove empty string ''
        # To handle the case when there is one extra comma: `CUDA_VISIBLE_DEVICES=0,1,2,3, python3 ...`
        available_gpus = [x for x in os.environ['CUDA_VISIBLE_DEVICES'].split(',') if x != '']
    except Exception:
        # If the env variable is not set, we use all GPUs
        available_gpus = [str(x) for x in range(torch.cuda.device_count())]
    n_gpus = len(available_gpus)
    procs_by_gpu = [None]*n_gpus

    while len(commands) > 0:
        for idx, gpu_idx in enumerate(available_gpus):
            proc = procs_by_gpu[idx]
            if (proc is None) or (proc.poll() is not None):
                # Nothing is running on this GPU; launch a command.
                cmd = commands.pop(0)
                new_proc = subprocess.Popen(
                    f'CUDA_VISIBLE_DEVICES={gpu_idx} {cmd}', shell=True)
                procs_by_gpu[idx] = new_proc
                break
        time.sleep(1)

    # Wait for the last few tasks to finish before returning
    for p in procs_by_gpu:
        if p is not None:
            p.wait()

REGISTRY = {
    'local': local_launcher,
    'dummy': dummy_launcher,
    'dummy_multi': dummy_multi_files_launcher,
    'multi_gpu': multi_gpu_launcher
}

try:
    from domainbed import facebook
    facebook.register_command_launchers(REGISTRY)
except ImportError:
    pass
